#ifdef __USAGE
%C - send I/O commands to a tape controller

%C -t Unit_type [-a io_port] -c	Command [-c Command]...
Options:
 -t         Type of Unit: Choices are: Archive, PS2, External, Internal
 -a         I/O port address of tape unit. Entered value is hexadecimal.
            Default is 0x300.
 -c         A command to execute. Multiple commands may be specified.
            e.g. -c rewind -c status -c erase -c rewind -c status
Available commands:
  reset      Reset the tape controller.
  rewind     Move to beginning of tape.
  erase      Move to beginning of tape and erase.
  writefm    Write a file mark at the current position.
  forward    Move forward to next file mark.
  status     Display tape status.
  test1      Self-test 1 (some controllers only)
  test2      Self-test 2 (some controllers only)

e.g. %C -t Archive -a 200 -c reset -c rewind
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <sys/types.h>

#include "tapedef.h"

char	*stest_byte2[] = {
	"Motor/Tach error\n",
	"Stal1 detected error\n",
	"Upper tape hole sensor error\n",
	"Lower tape hole sensor error\n",
	"Cartridge_IN sensor error\n",
	"Unsafe sensor error\n",
	"",  	/* byte 2 & 3 defs -> 0 = read after write test, 1 = read test */
	""		/* set to 1 if self test passed */
	};

char *stest_byte3[] = {
	"Track 0 error; excessive error count in R_A_W\n",
	"Track 1 error; excessive error count in R_A_W\n",
	"Stepper error (during READ test)\n",
	"Block address error (during R_A_W or READ test)\n",
	"LSI controller chip error\n",
	"Buffer error (16K RAM)\n",
	"Data seperator error\n",
	"8155 (PIA) test failed\n"
	};

#define MAX_COMMANDS 10  /* total commands allowed */
#define AVA_COMMANDS  7  /* available commands */
char *commands[] = {     /* commands to the tape unit */
    "reset",
    "writefm",
    "forward",
    "rewind",
    "erase",
    "status",
    "test1",
    "test2"
    };

void main(int argc, char *argv[]) {
    int opt;
    char tapetype, foundit;
    int commandstodo[MAX_COMMANDS];
    int commandcount = 0;
    int i;

	while ((opt = getopt(argc,argv,"t:a:c:")) != -1) {
		switch(opt) {
			case 't': tapetype = optarg[0];
					  break;
			case 'a': io_port = strtol( optarg,NULL,16);
					  break;
			case 'c': /* find a command */
                      foundit = 0;
                      for ( i = 0; i < AVA_COMMANDS && !foundit; i++ ) { 
                          if ( strncmp(commands[i],optarg,strlen(commands[i])) == 0 ) {
                             commandstodo[commandcount++] = i;
                             foundit = 1;
                             }
                          if ( commandcount >= MAX_COMMANDS ) {
                             fprintf(stderr,"Maximum of %d commands at one time.\n",MAX_COMMANDS);
                             exit(-1);
                             }
                          }
                      if ( !foundit ) { /* we never found the command */
                         fprintf(stderr,"Unrecognizable command '%s'\n",optarg);
                         exit(-1);
                         }
					  break;
			default:  break;
		}
	}

    switch ( tapetype ) {
        case 'A':
        case 'a': /* archive */
                  PS2 = 0;
	              unit_type = 'a';
                  CTRL_PORT = STATUS_PORT = io_port + 1;
		          CMD_PORT = DATA_PORT = io_port;
		          S_RDY = 0x40; S_EXC = 0x20; ONL = 0x00; RESET = 0x80; REQ = 0x40;
                  break;
        case 'E':
        case 'e': /* external */
                  PS2 = 0;
	              unit_type = 'e';
		          CTRL_PORT = STATUS_PORT = io_port;
		          CMD_PORT  = DATA_PORT = io_port + 1;
		          S_RDY = 0x01; S_EXC = 0x02; ONL = 0x01; RESET = 0x02; REQ = 0x04;
                  break;
        case 'I':
        case 'i': /* internal */
                  PS2 = 0;
	              unit_type = 'i';
		          CTRL_PORT = STATUS_PORT = io_port;
		          CMD_PORT  = DATA_PORT = io_port + 1;
		          S_RDY = 0x01; S_EXC = 0x02; ONL = 0x01; RESET = 0x02; REQ = 0x04;
                  break;
        case 'P':
        case 'p': /* PS/2 */
                  PS2 = 1;
	              unit_type = 'e';
		          CTRL_PORT = STATUS_PORT = io_port;
		          CMD_PORT  = DATA_PORT = io_port + 1;
		          S_RDY = 0x01; S_EXC = 0x02; ONL = 0x01; RESET = 0x02; REQ = 0x04;
                  break;
        default : /* none specified */
                  fprintf(stderr,"You must specify a tape unit type (-t)\n");
                  exit(-1);
		}

    if ( io_in(STATUS_PORT) == 0xff ) {
       fprintf(stderr,"No controller found at port 0x%x\n",io_port);
       exit(-1);
       }

    for ( i=0; i < commandcount; i++ ) {
        switch( commandstodo[i] ) {
            case 0: /* reset */
                    tape_reset();
                    break;
            case 1: /* writefm */
                    tape_write_fm();
                    break;
            case 2: /* forward (to next file mark */
                    tape_read_fm();
                    break;
            case 3: /* rewind */
                    tape_simple_cmd(TREWIND);
                    break;
            case 4: /* erase */
                    tape_reset();
                    tape_simple_cmd(TERASE);
                    break;
            case 5: /* status */
                    report_status(tape_read_status());
                    break;
            case 6: /* test1 */
                    run_self_test(1);
                    break;
            case 7: /* test2 */
                    run_self_test(2);
                    break;
            }
         }
}

tape_reset() {
	int i;

	online = 0;
	io_out(CTRL_PORT, RESET);			/* Reset line TRUE */
	pausetape(25);							/* Wait at least 25 usec */
	wait_till_not_rdy();				/* Wait for NOT ready */
	while(isexc() == 0)					/* Wait for exception */
		JMPNXT;
	io_out(CTRL_PORT, online);			/* Reset line FALSE */

	if((i = tape_read_status()) != POWER_RST)	/* Clear the status bytes */
		report_status(i);				/* Print any errors */
	}

tape_simple_cmd(cmd)
unsigned char cmd;
{
	int status = 0;

	wait_till_rdy();
	io_out(CTRL_PORT, online);
	io_out(CMD_PORT, last_cmd = cmd);
	io_out(CTRL_PORT, REQ | online);
	wait_till_rdy();
	io_out(CTRL_PORT, online);
	pausetape(100);					/* Wait at least 100 usec */

	for(;;) {
		if(isrdy())
			break;
		if(isexc()) {
			status = tape_read_status();
			report_status(status);
			break;
		}
	}
	return(status);
}

wait_till_rdy() {
	while(isrdy() == 0)
		JMPNXT;
	}

wait_till_not_rdy() {
	while( isrdy() )
		JMPNXT;
	}

isrdy() {
	return((io_in(STATUS_PORT) & S_RDY) == 0);		/* Bit 0 when ready */
	}

isexc() {
	return((io_in(STATUS_PORT) & S_EXC) == 0);		/* Bit 0 when execption */
	}

/*
 * pausetape for at least the passed number of usec. We actually pausetape
 * a fair bit longer to play it safe.
 */

void pausetape(int usecs)
{
int ios;

usecs *=2;
ios = usecs + usecs/4 + ((usecs & 3) != 0);
    while (ios--) {
        inb(0);
    }
}

tape_read_status() {
	int i;

	io_out(CMD_PORT, last_cmd = TREAD_STAT);/* Put command on the bus */
	io_out(CTRL_PORT, REQ | online);		/* Set request and online */
	wait_till_rdy();
	io_out(CTRL_PORT, online);				/* Reset request but keep online */
	/*
	 * At this point the ready line will go from ready to not-ready
	 * It can take 100 usec before it goes not ready.
	 */
	pausetape(100);								/* Wait at least 100 usec */
	for(i = 0 ; i < 6 ; ++i) {
		wait_till_rdy();
		srb[i] = io_in(DATA_PORT);			/* Get the status */
		io_out(CTRL_PORT, REQ + online);	/* Set request */
		wait_till_not_rdy();
		io_out(CTRL_PORT, online);			/* Reset request */
		}

	wait_till_rdy();
	return(map_status());
	}

map_status() {
	unsigned errnum = ((srb[0] << 8) & 0xff00) | (srb[1] & 0xff) | 0x8000;

	if(srb[1] == 0  &&  (srb[0] & 0x80)) {
		if(srb[0] & 0x20)
			errnum = NO_DRIVE;
		else if(srb[0] & 0x40)
			errnum = NO_CART;
		else if(srb[0] & 0x01)
			errnum = RD_FILEMARK;
		else if(srb[0] & 0x08)
			errnum = EOM;
		else if(srb[0] & 0x10)
			errnum = WRT_PROTECT;
		else if(srb[0] & 0x06)
			errnum = (srb[0] & 0x04) ? RD_FBXFER : RD_BBXFER;
		}
		
	if(srb[1] & 0x80) {
		if(srb[1] & 0x01)
			errnum = POWER_RST;
		else if((srb[0] & 0x90) == 0x90  &&  (srb[1] & 0x77) == 0)
			errnum = WRT_PROTECT;
		else if(srb[1] & 0x08)
			errnum = (srb[0] & 0x84) == 0x84 ? RD_WRT_ABORT : BOM;
		else if(srb[1] & 0x20)
			errnum = (srb[0] & 0x08) ? RD_EOM : RD_NODATA;
		else if(srb[1] & 0x40)
			errnum = ILLEGAL;
		else if(srb[1] & 0x10)
			errnum = MARGINAL;
		else if(srb[0] & 0x08)
			errnum = EOM;
		}
#ifdef DEBUG
fprintf(stderr, "Mapping status bytes %2.2x %2.2x --> %4.4x\n",
					srb[0] & 0xff, srb[1] & 0xff, errnum);
#endif
	return(errnum);
	}

report_status(error)
unsigned error;
	{
	register char *p = NULL;
	static char *error_strs[] ={
		"aOK",
		"qNo cartrige in drive",
		"qDrive is off or disconnected",
		"aCartridge is write protected",
		"cEnd of tape",
		"aRead or write abort has occurred",
		"sRead error, bad block transfer",
		"sRead error, filler block transfer",
		"sNo data detected on this section of tape",
		"cEnd of tape on read",
		"sFilemark was read",
		"aTape unit reports that an illegal command was attempted",
		"cPower on reset completed",
		"cMarginal block detected",
		} ;	/* a - abort  c - continue  i = ignore  q - query  s - stop */

	if(error < sizeof(error_strs)/sizeof(char *))
		p = error_strs[error];
	else
		p = error_strs[0];

	fprintf(stderr, "%s (%4.4X).\n", p + 1, error);

#ifdef DEBUG
    fprintf(stderr,"Keep going....hit a key");
    getchar();
#endif
	}

tape_read_fm() {
	online = ONL;
	return(tape_simple_cmd(TREADFM));
	}

tape_write_fm() {
	online = ONL;
	return(tape_simple_cmd(TWRITEFM));
	}

run_self_test(testno)
char	testno;
{
	int i;
	unsigned char c;	
	char	*p;

	if (testno == 2)   {
		fprintf(stderr,"\n** WARNING - this test will destroy all data on the tape.\n");
		fprintf(stderr,"Type 'Y' to continue... ");
		tcflush(0,TCIFLUSH);
		if ( getchar() != 'Y' )
           return(1);
	}

	fprintf(stderr,"\nRunning self test %02d ...\n\n", testno);

	tape_reset();
	tape_simple_cmd((testno==1) ? TSELFTST1 : TSELFTST2);
	/*
	*	Bit 5 of byte 0 indicates error (0=OK, 1=ERROR)
	*/
	if (!(srb[0] & 0x20))	{
		fprintf(stderr,"\07No errors detected.\n");
		return(0);
	}

	fprintf(stderr,"\07Errors found :\n");

	for(p=&srb[0], i=0; i < 6; i++)
		fprintf(stderr," %02uh", (unsigned char) *p++); 

	c = srb[2];
	for (i=0; i < 8; p++, i++)	{
		if (c & 0x01)
			fprintf(stderr,"%s", stest_byte2[i]);
		c = c >> 1;
	}
	/*
	*	Status byte 3 only valid on self test 1
	*/
	if (testno == 1)	{
		c = srb[3];
		for (i=0; i < 8; p++, i++)	{
			if (c & 0x01)
				fprintf(stderr,"%s", stest_byte3[i]);
			c = c >> 1;
		}
	}
	/*
	*	Status byte 4 & 5 only valid on self test 2
	*/
	if (testno == 2)
		fprintf(stderr,"Data error count = %04d\n", (srb[4]<<8) | srb[5]);
}

